/**
 * Copyright (c) FOM-Nikhef 2014-
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *
 *  Authors:
 *  2014-
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */

#include <lcmaps/lcmaps_log.h>
#include <openssl/x509v3.h>
#include <openssl/pem.h>
#include <string.h>
#include <sys/types.h>
#include <regex.h>

#include "lcmaps_subuserdn.h"


/************************************************************************
 * defines
 ************************************************************************/

/* OID of RFC 3820 proxy */
#define OID_RFC_PROXY	    "1.3.6.1.5.5.7.1.14"

/* OID of Robot cert (non-human automated client or robot) */ 
#define OID_ROBOT_CERT	    "1.2.840.113612.5.2.3.3.1"

/* Regex pattern for robot EEC */
#define ROBOT_REGEX	    ".*/CN=[rR]obot[^[:alnum:]]+"

/* Buffer size for converting objects to strings */
#define OBJ_BUF_SIZE	    80


/************************************************************************
 * private prototypes
 ************************************************************************/
 
/* Returns wether certificate has given policy O.I.D.
 * return 1 for success, 0 for failure */
static int cert_has_oid(X509 *cert, const char *oid);

/* Returns wether certificate has robot DN
 * return 1 on success, 0 on no match or error */
static int cert_has_robot_dn(const char *dn);

/* Returns wether certificate is RFC3820 proxy certificate
 * return 1 when proxy is RFC, 0 otherwise */
static int proxy_is_rfc(X509 *proxy);


/************************************************************************
 * public functions
 ************************************************************************/

/**
 * Convert PEM string in certstring to chain of X509, returned in certstack.
 * Function is identical to lcmaps_pem_string_to_x509_chain()
 * return 0 on success, non-zero on error
 */
int pem_string_to_x509_chain(STACK_OF(X509) **certstack, char *certstring)
{
    STACK_OF(X509) *mystack=NULL;
    BIO *certbio = NULL;
    STACK_OF(X509_INFO) *sk=NULL;
    X509_INFO *xi;

    /* protect input */
    if (certstack==NULL || certstring==NULL)	{
	lcmaps_log(LOG_ERR,"%s: certstack and/or certstring is NULL\n",
		__func__);
	return -1;
    }

    /* initialize new X509 chain and memory BIO */
    if ( (mystack = sk_X509_new_null()) == NULL ||
	 (certbio = BIO_new_mem_buf(certstring, -1)) == NULL)
    {
	lcmaps_log(LOG_ERR,"%s: out of memory\n", __func__);
	return -1;
    }

    /* Convert PEM via memory BIO to stack of X509_INFO */
    sk=PEM_X509_INFO_read_bio(certbio, NULL, NULL, NULL);
    BIO_free(certbio);
    if (!sk)
	goto err;

    /* Loop over the stack and convert */
    while (sk_X509_INFO_num(sk)) {
	xi=sk_X509_INFO_shift(sk);
	if (xi->x509 != NULL) {
	    sk_X509_push(mystack, xi->x509);
	    xi->x509=NULL;
	}
	X509_INFO_free(xi);
    }
    sk_X509_INFO_free(sk);

    /* Check there is at least one certificate */
    if (!sk_X509_num(mystack))
	goto err;

    /* Cleanup and return success */
    *certstack=mystack;

    return 0;

err:
    /* Cleanup and error out */
    sk_X509_pop_free(mystack, X509_free);
    return 1;
}


/**
 * Looks for the first proxy DN in given certificate chain, iff the EEC is a
 * robot certificate.
 * Return 0 on success, -1 on failure
 */
int lcmaps_get_subuser_dn(char **account, STACK_OF(X509) *chain)	{
    char *dn=NULL;
    X509 *cert;
    int i, depth, amount_of_CAs=0, eec_index;
    const int purpose=X509_PURPOSE_MIN + X509_PURPOSE_get_by_sname("sslclient");

    /* protect input */
    if (chain==NULL)	{
	lcmaps_log(LOG_ERR, "%s: invalid chain\n", __func__);
	return -1;
    }

    /* Go through chain to look for EEC */
    depth = sk_X509_num(chain);

    /* How many CA certs are there in the chain? */
    for (i = 0; i < depth; i++) {
	/* final argument to X509_check_purpose() is whether to check for CAness
	 * */
	if (X509_check_purpose(sk_X509_value(chain, i), purpose, 1))
	    amount_of_CAs++;
    }

    /* Getting the user cert (filtering the CAs and such) */
    eec_index = depth-(amount_of_CAs+1);

    /* Next check also catches depth==0 */
    if (eec_index < 0) {
	/* Counting the stack went wrong. Erroneous stack? */
	lcmaps_log(LOG_WARNING,
		"%s: cannot parse certificate chain: "
		"depth-(amount_of_CAs+1)=%d < 0\n", __func__, eec_index);
	return -1;
    }

    /* Do we have a proxy cert after the EEC? */
    if (eec_index==0) {
	lcmaps_log(LOG_NOTICE,
		"%s: cannot find proxy cert in chain, EEC seems end\n",
		__func__);
	return -1;
    }

    /* We got the EEC */
    cert=sk_X509_value(chain, eec_index);

    /* Get EEC DN, note: we cannot use X509_NAME_print_ex() since it cannot
     * print the / delimiter notation. */
    dn=X509_NAME_oneline(X509_get_subject_name(cert), NULL, 0);

    /* Get DN and oids */
    if (!cert_has_oid(cert, OID_ROBOT_CERT))	{
	lcmaps_log(LOG_INFO,
		"%s: EEC does not contain robot OID %s, using DN...\n",
		__func__, OID_ROBOT_CERT);
	if (!cert_has_robot_dn(dn))	    {
	    lcmaps_log(LOG_NOTICE, "%s: DN is not a valid robot DN: %s \n",
		    __func__, dn);
	    free(dn);
	    return -1;
	}
    }

    /* Log the DN */
    lcmaps_log(LOG_DEBUG, "%s: Found robot EEC with DN: %s\n", __func__, dn);
    free(dn);

    /* First proxy should be the next */
    cert=sk_X509_value(chain, eec_index-1);

    /* Get proxy dn */
    dn=X509_NAME_oneline(X509_get_subject_name(cert), NULL, 0);

    /* Check it is an RFC proxy */
    if (!proxy_is_rfc(cert))	{
	lcmaps_log(LOG_NOTICE,
		"%s: Proxy is NOT an RFC3820 proxy, DN: %s\n", __func__, dn);
	free(dn);
	return -1;
    }

    /* Log on log debug */
    lcmaps_log(LOG_INFO, "%s: Found robot cert with RFC proxy with DN: %s\n",
	    __func__, dn);

    /* copy dn into account */
    *account = dn;

    return 0;
}


/************************************************************************
 * private functions
 ************************************************************************/

/**
 * Checks whether certificate contains given OID
 * return 1 for success, 0 for failure
 */
static int cert_has_oid(X509 *cert, const char *oid)   {
    char buffer[OBJ_BUF_SIZE];
    CERTIFICATEPOLICIES *policy_oids;
    POLICYINFO *pol=NULL;
    int i,numpols,rc=0;

    /* Basic sanity check */
    if (cert==NULL || oid==NULL)
	return 0;

    /* Get Certificate Policy OIDs */
    policy_oids = X509_get_ext_d2i(cert, NID_certificate_policies, 0, 0);
    if (!policy_oids)
	return 0;

    /* Get number of OIDs */
    numpols = sk_POLICYINFO_num(policy_oids);
    if (numpols==0) {
	CERTIFICATEPOLICIES_free(policy_oids);
	return 0;
    }

    /* Look for given oid */
    for (i=0; i<numpols; i++)   {
	pol = sk_POLICYINFO_value(policy_oids, i);
	/* OBJ_obj2txt(3ssl) says that 80 should be more than enough */
	OBJ_obj2txt(buffer, OBJ_BUF_SIZE, pol->policyid, 0);
	if (strcmp(buffer, oid)==0) {
	    rc=1;
	    break;
	}
    }

    /* Free memory */
    CERTIFICATEPOLICIES_free(policy_oids);

    return rc;
}


/**
 * Checks whether DN is of robot form i.e. matches ROBOT_REGEX
 * return 1 on success, 0 on no match or error
 */
static int cert_has_robot_dn(const char *dn)	{
    const char *logstr="lcmaps_subuserdn-cert_has_robot";
    regex_t preg;
    char buffer[1024];
    int rc;

    /* Compile pattern */
    if ( (rc=regcomp(&preg, ROBOT_REGEX, REG_NOSUB)) )    {
	regerror(rc, &preg, buffer, 1024);
	lcmaps_log(LOG_ERR, "%s: regcomp failed: %s\n", logstr, buffer);
	regfree(&preg);
	return 1;
    }

    /* Do comparison and free memory */
    rc=regexec(&preg, dn, 0, NULL, 0);
    regfree(&preg);

    switch(rc)	{
	case 0:
	    return 1;
	case REG_NOMATCH:
	    return 0;
	default:
	    lcmaps_log(LOG_ERR,
			"%s: unexpected return value %d for regexec\n",
			logstr, rc);
	    return 0;
    }
}

/**
 * Checks whether proxy is RFC3820 compliant i.e. has OID_RFC_PROXY proxy
 * extension OID.
 * return 1 when proxy is RFC, 0 otherwise
 */
static int proxy_is_rfc(X509 *proxy)	{
    int ext_count, i, rfc=0;
    X509_EXTENSION *ex;
    char buffer[OBJ_BUF_SIZE];

    /* Loop of all certificate extensions */
    ext_count=X509_get_ext_count(proxy);
    for (i = 0; i < ext_count; i++) {
        ex = X509_get_ext(proxy, i);
        if (X509_EXTENSION_get_object(ex)) {
            OBJ_obj2txt(buffer, OBJ_BUF_SIZE, X509_EXTENSION_get_object(ex), 1);
	    /* Compare found extension if RFC_PROXY OID */
            if (strcmp(buffer, OID_RFC_PROXY) == 0)	{
		rfc=1;
		break;
	    }
	}
    }

    return rfc;
}

